# Ruby port of the PHP flowing block module for FPDF by Damon Kohler 
# the Ruby port is a bit enhanced and supports colored fonts as well
#
# original code by Damon Kohler: 
# http://www.fpdf.de/downloads/addons/65/
#
# Ruby port by Martin Müller: 
# http://www.martinonrails.com
# locomotive@martinonrails.com
#
# license: GPL
#

require 'fpdf'


class PDF < FPDF
  
    def initialize(orientation='L', unit='mm', format='A6')
        super(orientation, unit, format)
    
        @tags = { 'B' => false, 'I' => false, 'U' => false }
        @href = ''
	
        @flowingBlockAttr = Hash.new
    end
    
    ## FLOWING_BLOCK EXTENSION
    ## ==============================================================================
    
    
    def getMaxHeight
    	@h
    end
    
    def getMaxWidth
    	@w
    end
    
    def getTextColor
    	@TextColor
    end
  
    def saveFont

        saved = Hash.new

        saved[ 'family' ] = @FontFamily
        saved[ 'style' ] = @FontStyle
        saved[ 'sizePt' ] = @FontSizePt
        saved[ 'size' ] = @FontSize
        saved[ 'curr' ] = @CurrentFont 
        saved[ 'color' ] = @TextColor

        saved

    end

    def restoreFont( saved )

        @FontFamily = saved[ 'family' ]
        @FontStyle = saved[ 'style' ]
        @FontSizePt = saved[ 'sizePt' ]
        @FontSize = saved[ 'size' ]
        @CurrentFont = saved[ 'curr' ]
        @TextColor = saved[ 'color' ]

        if( @page > 0)
    	
    		s = sprintf( 'BT /F%d %.2f Tf ET', @CurrentFont[ 'i' ], @FontSizePt )
            
    		out (s)
            
	    end

    end

    def newFlowingBlock( w, h, b = 0, a = 'J', f = 0 )

        # cell width in points
    	
        @flowingBlockAttr[ 'width' ] =  w * @k

        # line height in user units
        @flowingBlockAttr[ 'height' ] = h

        @flowingBlockAttr[ 'lineCount' ] = 0

        @flowingBlockAttr[ 'border' ] = b
        @flowingBlockAttr[ 'align' ] = a
        @flowingBlockAttr[ 'fill' ] = f

        @flowingBlockAttr[ 'font' ] = Array.new
        @flowingBlockAttr[ 'content' ] = Array.new
        @flowingBlockAttr[ 'contentWidth' ] = 0

    end

    
    def finishFlowingBlock()

        maxWidth = @flowingBlockAttr[ 'width' ]
        lineHeight = @flowingBlockAttr[ 'height' ]

        border = @flowingBlockAttr[ 'border' ]
        align = @flowingBlockAttr[ 'align' ]
        fill = @flowingBlockAttr[ 'fill' ]

        content = @flowingBlockAttr[ 'content' ]
        font = @flowingBlockAttr[ 'font' ]  

        # set normal spacing
        out( sprintf( '%.3f Tw', 0 ) )

        # print out each chunk

        # the amount of space taken up so far in user units
        usedWidth = 0

        content.each_with_index do | chunk, k |

            b = ''

            unless border.index('B').nil? 
                b = b + 'B'
            end
              
            if k == 0 && !( border.index('L').nil?  )  
                b = b +'L'
            end

            if ( k == content.size - 1 ) && !( border.index('R').nil?  ) 
                b =  b + 'R'
            end

            restoreFont( font[ k ] )
      
            # if it's the last chunk of this line, move to the next line after
      
            if ( k == content.size - 1 )  #CHECK was: if ( $k == count( $content ) - 1 )
                Cell( ( maxWidth / @k ) - usedWidth + 2 * @cMargin, lineHeight, chunk, b, 1, align, fill)      # $this->Cell( ( $maxWidth / $this->k ) - $usedWidth + 2 * $this->cMargin, $lineHeight, $chunk, $b, 1, $align, $fill );
            else
                Cell( GetStringWidth( chunk ), lineHight, chunk, b, 0, align, fill)     
            end
      
            usedWidth = usedWidth + GetStringWidth( chunk )   

        end

        @flowingBlockAttr[ 'width' ] = maxWidth 
        @flowingBlockAttr[ 'height' ] = lineHeight

        @flowingBlockAttr[ 'border' ] = border 
        @flowingBlockAttr[ 'align' ] = align 
        @flowingBlockAttr[ 'fill' ] = fill

        @flowingBlockAttr[ 'content' ] = content 
        @flowingBlockAttr[ 'font' ]   = font

    end
      
      
    def WriteFlowingBlock( s )

        # width of all the content so far in points
        contentWidth =  @flowingBlockAttr[ 'contentWidth' ]
        
        # cell width in points
        maxWidth = @flowingBlockAttr[ 'width' ]

        lineCount = @flowingBlockAttr[ 'lineCount' ]

        # line height in user units
        lineHeight = @flowingBlockAttr[ 'height' ]

        border = @flowingBlockAttr[ 'border' ]
        align = @flowingBlockAttr[ 'align' ]
        fill = @flowingBlockAttr[ 'fill' ]

        content = @flowingBlockAttr[ 'content' ]
        font = @flowingBlockAttr[ 'font' ]

        font.push(saveFont())    
        
        content = content.push('')

        currContent = content[ content.size - 1 ]    
        currContentLastFlag = content.size - 1	

        # where the line should be cutoff if it is to be justified
        cutoffWidth = contentWidth;

        # for every character in the string
        0.upto( s.length - 1 ) do |i|    

        # extract the current character
    		c = s[i,1][0]   

    		# get the width of the character in points
		    cw = @CurrentFont[ 'cw' ][ c ] * ( @FontSizePt.to_f / 1000 )   

		    if ( c.chr == ' ' ) 

		        currContent = currContent + ' '
		        cutoffWidth = contentWidth

		        contentWidth = contentWidth + cw

		        next # CHECK was:continue;

		    end

		    # try adding another char
		    if ( contentWidth + cw > maxWidth )

		        # won't fit, output what we have
		        lineCount = lineCount + 1

		        # contains any content that didn't make it into this print
		        savedContent = ''
		        savedFont = Array.new

		        # first, cut off and save any partial words at the end of the string
		        words = currContent.split     

		        # if it looks like we didn't finish any words for this chunk
		        if ( words.size == 1 )

		            # save and crop off the content currently on the stack
		            savedContent = content.pop
		            savedFont = font.pop

		            # trim any trailing spaces off the last bit of content
		            currContent = content[ content.size - 1 ]
		            currContentLastFlag = content.size - 1 		
		            

		            currContent.rstrip!         

		        else
		            # otherwise, we need to find which bit to cut off
		            lastContent = ''

		            0.upto( words.size - 2 ) do |w|    # for ( $w = 0; $w < count( $words ) - 1; $w++)
		                lastContent =  lastContent + words[w] + " "    
		            end

			        savedContent =  words[ words.size - 1 ]       
			        savedFont = saveFont()

			        # replace the current content with the cropped version
			        currContent = lastContent.rstrip
			        
		        	if ( currContentLastFlag > 0 )
		        		content[ currContentLastFlag ] = currContent
		        	else
		        		content[ 0 ] = currContent
		        	end


		        end

		        # update $contentWidth and $cutoffWidth since they changed with cropping
		        contentWidth = 0 
		    
		        content.each_with_index do | chunk, k |  

			        restoreFont( font[ k ] )
			        contentWidth = contentWidth + GetStringWidth( chunk ) * @k;

		        end

		        cutoffWidth = contentWidth;

                # if it's justified, we need to find the char spacing
		        if( align == 'J' )
			        # count how many spaces there are in the entire content string
			        numSpaces = 0

			        content.each do | chunk |
			            numSpaces = numSpaces +  chunk.length - chunk.gsub(/ /, '').length  
			        end

			        # if there's more than one space, find word spacing in points
			        if ( numSpaces > 0 )
			            @ws = ( maxWidth - cutoffWidth ) / numSpaces    
			        else
			            @ws = 0
			        end

			        out( sprintf( '%.3f Tw', @ws ) )

		        else
		            # otherwise, we want normal spacing
			        out( sprintf( '%.3f Tw', 0 ) )
		        end

		        # print out each chunk
		        usedWidth = 0
		    
		        content.each_with_index do | chunk, k |

			        restoreFont( font[ k ] )

			        stringWidth = GetStringWidth( chunk ) + ( @ws * ( chunk.length - chunk.gsub(/ /, '').length ) / @k );  

			        # determine which borders should be used
			        b = ''
			        
			        border = border.to_s

                    if ( lineCount == 1 ) && !(border.index('T').nil? )   
                        b = b + 'T'
                    end
                        
                    if ( k == 0 ) && !(border.index('L').nil? )           
                        b = b + 'L'
                    end

                    if ( k == content.size - 1 ) && !(border.index('R').nil? )       
                        b = b +'R'
                    end

			        # if it's the last chunk of this line, move to the next line after
			        if ( k == content.size  - 1 )
			            Cell( ( maxWidth / @k ) - usedWidth + 2 * @cMargin, lineHeight, chunk, b, 1, align, fill )  
			        else
			            Cell( stringWidth + 2 * @cMargin, lineHeight, chunk, b, 0, align, fill )
			            @x = @x - 2 * @cMargin
			        end

			        usedWidth = usedWidth + stringWidth

                end

		        # move on to the next line, reset variables, tack on saved content and current char
		        restoreFont( savedFont )

                font = Array.new
                font.push(savedFont)         
		        
		        content = Array.new
		        content.push(savedContent + s[i,1])

		        currContent = content[ 0 ]
				currContentLastFlag = 0		

		        contentWidth = GetStringWidth( currContent ) * @k
		        cutoffWidth = contentWidth


		        # another character will fit, so add it on
		    else

		        contentWidth = contentWidth + cw  
		        toadd = s[i,1] 
		        currContent = currContent.to_s + toadd.to_s

		    end

		end
	    
        if ( currContentLastFlag > 0 )
        	content[ currContentLastFlag ] = currContent
        else
        	content[ 0 ] = currContent
        end
	    
        # width of all the content so far in points
        @flowingBlockAttr[ 'contentWidth' ] = contentWidth 
        
        # cell width in points
        @flowingBlockAttr[ 'width' ] = maxWidth 

        @flowingBlockAttr[ 'lineCount' ] = lineCount 

        # line height in user units
        @flowingBlockAttr[ 'height' ] = lineHeight 

        @flowingBlockAttr[ 'border' ] = border.to_s
        @flowingBlockAttr[ 'align' ] = align 
        @flowingBlockAttr[ 'fill' ] = fill 

        @flowingBlockAttr[ 'content' ] = content
        @flowingBlockAttr[ 'font' ] = font
	    
    end
end   
